"""
This module provides tools for working with PSF files.

"""

from pytopol.utils import build_res_chain, build_pairs
from pytopol.pdb import PDBSystem
from pytopol import blocks

import os
import logging
import time

# create logger
module_logger = logging.getLogger('mainapp.psf')



class PSFSystem(blocks.System):
    def __init__(self, psffile):
        """ Initialization of a PSF file.

        Args:
            psffile
                the path to the psf file
            pdbfile
                optional, the path to the pdb file

        Attributes:
            self.lgr        : logger.Logger
            self.psfile     : str, path to the psf file
            self.molecules  : a tuple of Molecule instances

        """

        self.lgr = logging.getLogger('mainapp.psf.PSFSystem')
        self.lgr.debug(">> entering PSFSystem")

        super(PSFSystem, self).__init__()

        self.psffile = psffile
        self.molecules = tuple([])

        # parse the psf file and create one molecule
        mol = self._parse(self.psffile)
        self.molecules = tuple([mol])

        self.lgr.debug("<< leaving PSFSystem")


    def __repr__(self):
        natoms = sum([len(m.atoms) for m in self.molecules])
        return 'PSF file with %6d atoms and %2d molecules' % (
            natoms, len(self.molecules))



    def add_pdbfile(self, pdbfile, mol):
        """ add coordinates form a pdb file to the system."""

        pdb = PDBSystem(pdbfile, guess_mols=False)

        if len(pdb.atoms) == len(mol.atoms):
            for i, atom in enumerate(mol.atoms):
                atom.coords = pdb.atoms[i].coords
            self.lgr.debug("coordinates from pdb file were added")
        else:
            self.lgr.error("the number of atoms in the pdb and psf files doesn't match: %6d vs %6d" % (
                len(pdb.atoms), len(mol.atoms)))



    def _parse(self, psffile):
        """ Parse a psf file.

        Args:
            psffile : str, path to the psf file
        Returns:
            a Molecule instance or False

        """

        self.lgr.debug("parsing psf file: %s" % psffile)

        t1 = time.time()

        # check if the file exists
        if not os.path.exists(psffile):
            self.lgr.critical("file doesn't exist")
            return False

        # initialize empty list for data
        mol = blocks.Molecule()

        # supported formats
        psf_formats = {
            'NAMD': {
                'sections': {
                    '!NATOM':  {'type':'atom',    'n':(9,11),'multiple':False,'func':self._atom_line},
                    '!NBOND':  {'type':'bond',    'n':2,'multiple':True, 'func':self._badi_line},
                    '!NTHETA': {'type':'angle',   'n':3,'multiple':True, 'func':self._badi_line},
                    '!NPHI':   {'type':'dihedral','n':4,'multiple':True, 'func':self._badi_line},
                    '!NIMPHI': {'type':'improper','n':4,'multiple':True, 'func':self._badi_line},
                    '!NCRTERM':{'type':'cmap',    'n':8,'multiple':False,'func':self._badi_line},
                },
            },
        }


        # find the psf format
        with open(psffile) as f:
            for line in f:
                psffmt = self._find_psf_format(line)
                break

        # check if the format is valid
        if psffmt is False:
            # assume the format is 'NAMD'
            psffmt = 'NAMD'
            #return False

        elif psffmt not in list(psf_formats.keys()):
            self.lgr.error("psf format '%s' is not supported" % (psffmt))
            return False

        # parse the file
        _sec = None
        known_sections = list(psf_formats[psffmt]['sections'].keys())

        with open(psffile) as f:
            for line in f:
                line = line.strip()

                if line == '':
                    continue

                if '!' in line:
                    if line.split()[1].strip(':') in known_sections:
                        _sec = line.split()[1].strip(':')
                    else:
                        self.lgr.debug("skipping section: '%s'" % line)
                        _sec = None
                    continue

                if _sec is not None:
                    _conf = psf_formats[psffmt]['sections'][_sec]
                    result = _conf['func'](psffmt, line, _conf, mol)
                    if result is False:
                        self.lgr.error("couldn't parse this line in '%s' section:\n  %s" % (_sec, line))
                        return False

        # build chain and residues
        build_res_chain(mol)
        build_pairs(mol)

        t2 = time.time()
        self.lgr.debug("parsing took %4.1f seconds" % (t2-t1))

        return mol



    def split_psf(self):
        """Convert a psf Molecule to multiple Molecules.

        Using this function only makes sense if the segments in the PSF file
            are not covalently bond (usually this the case).

        Args:
           temp_mol : a psf Molecule instance

        Returns:
           list of Molecules or False

        """

        assert len(self.molecules) == 1
        temp_mol = self.molecules[0]

        self.lgr.debug("converting psf to multiple molecules based on chains")

        unique_chains = set([chain.name for chain in temp_mol.chains])
        if len(unique_chains) != len(temp_mol.chains):
            self.lgr.error("the name of the chains is not unique")
            return False


        # counter for different elements in the psf file
        _NA= _B = _A = _D = _I = _C  = _NP = 0
        molecules = []


        # Atom:Chain.name dictionary for easy lookup
        _AC_map = {}
        for atom in temp_mol.atoms:
            _AC_map[atom] = atom.residue.chain.name


        for chain in temp_mol.chains:
            m = blocks.Molecule()
            chainname = chain.name

            for res in chain.residues:
                for atom in res.atoms:
                    m.atoms.append(atom)
                    _NA += 1

            for b in temp_mol.bonds:
                if _AC_map[b.atom1] == chainname and _AC_map[b.atom2] == chainname:

                    m.bonds.append(b)
                    _B += 1

            for a in temp_mol.angles:
                if _AC_map[a.atom1] == chainname and _AC_map[a.atom2] == chainname and \
                   _AC_map[a.atom3] == chainname:

                    m.angles.append(a)
                    _A += 1

            for d in temp_mol.dihedrals:
                if _AC_map[d.atom1] == chainname and _AC_map[d.atom2] == chainname and \
                   _AC_map[d.atom3] == chainname and _AC_map[d.atom4] == chainname:

                    m.dihedrals.append(d)
                    _D += 1

            for i in temp_mol.impropers:
                if _AC_map[i.atom1] == chainname and _AC_map[i.atom2] == chainname and \
                   _AC_map[i.atom3] == chainname and _AC_map[i.atom4] == chainname:

                    m.impropers.append(i)
                    _I += 1


            for c in temp_mol.cmaps:
                if _AC_map[c.atom1] == chainname and _AC_map[c.atom2] == chainname and \
                   _AC_map[c.atom3] == chainname and _AC_map[c.atom4] == chainname and \
                   _AC_map[c.atom5] == chainname and _AC_map[c.atom6] == chainname and \
                   _AC_map[c.atom7] == chainname and _AC_map[c.atom8] == chainname:

                    m.cmaps.append(c)
                    _C += 1

            for p in temp_mol.pairs:
                if _AC_map[p.atom1] == chainname and _AC_map[p.atom2] == chainname:

                    m.pairs.append(p)
                    _NP += 1



            build_res_chain(m)
            m.renumber_atoms()
            molecules.append(m)

        # make sure we used all the enteties in the temp_mol
        assert len(temp_mol.atoms)     == _NA
        assert len(temp_mol.bonds)     == _B
        assert len(temp_mol.angles)    == _A
        assert len(temp_mol.dihedrals) == _D
        assert len(temp_mol.impropers) == _I
        assert len(temp_mol.cmaps)     == _C
        assert len(temp_mol.pairs)     == _NP

        self.molecules = tuple(molecules)



    def _find_psf_format(self, first_line):
        """Find the PSF format.

        Args:
            first_line: str, the first line in the psf file

        Returns:
            'NAMD' or False

        """
        _line = first_line.strip()
        if not _line.startswith('PSF'):
            self.lgr.error("the first line of psf file doesn't start with 'PSF'")
            return False
        else:
            fields = _line.split()
            if 'NAMD' in fields:
                return 'NAMD'
            else:
                self.lgr.warning("could not find NAMD keywork in "
                                 "the first line in the psf file")
                return False



    def _atom_line(self, psffmt, line, conf, m):
        """Parse an ATOM line in the psf file.

       Args:
           psffmt: str, 'NAMD'
           line:  str, a psf ATOM line
           conf: dict
           m:  Molecule

        """
        if psffmt == 'NAMD':
            f = line.split()
            if len(f) not in conf['n']:
                self.lgr.error("(e) the number of elements in atom line is '%d, expected to be %d" % (
                    len(f), conf['n']))
                return False

            if len(f) == 9:
                atnumb, segname, resnumb, resname, atname, attype, charge, mass, tmp = f
            elif len(f) == 11:
                atnumb, segname, resnumb, resname, atname, attype, charge, mass, tmp1, tmp2, tmp3 = f
            else:
                raise NotImplementedError
            a = blocks.Atom()
            a.name      = atname
            a.number    = int(atnumb)
            a.atomtype  = attype
            a.chain     = segname
            a.resname   = resname
            a.resnumb   = int(resnumb)
            a.charge    = float(charge)
            a.mass      = float(mass)

            m.atoms.append(a)

            return True

        else:
            raise NotImplementedError



    def _badi_line(self, psffmt, line, conf, m):

        if psffmt == 'NAMD':

            f = line.split()

            # check the number of elements in the line
            if conf['multiple']:
                assert len(f) % conf['n'] == 0
            else:
                assert len(f) == conf['n']

            # read the elements
            for i in range(int(len(f)/conf['n'])):
                if conf['type'] == 'bond':
                    _m = i * conf['n']
                    _n = i * conf['n'] + 1

                    b = blocks.Bond()
                    atom1 = m.anumb_to_atom(int(f[_m]))
                    atom2 = m.anumb_to_atom(int(f[_n]))

                    if atom1 and atom2:
                        b.atom1 = atom1
                        b.atom2 = atom2

                        m.bonds.append(b)
                    else:
                        raise TypeError

                elif conf['type'] == 'angle':
                    _m = i * conf['n']
                    _n = i * conf['n'] + 1
                    _o = i * conf['n'] + 2

                    a = blocks.Angle()
                    atom1 = m.anumb_to_atom(int(f[_m]))
                    atom2 = m.anumb_to_atom(int(f[_n]))
                    atom3 = m.anumb_to_atom(int(f[_o]))

                    if atom1 and atom2 and atom3:
                        a.atom1 = atom1
                        a.atom2 = atom2
                        a.atom3 = atom3

                        m.angles.append(a)
                    else:
                        raise TypeError

                elif conf['type'] in ('dihedral', 'improper'):
                    _m = i * conf['n']
                    _n = i * conf['n'] + 1
                    _o = i * conf['n'] + 2
                    _p = i * conf['n'] + 3

                    if conf['type'] == 'dihedral':
                        d = blocks.Dihedral()
                        cont = m.dihedrals
                    else:
                        d = blocks.Improper()
                        cont = m.impropers

                    atom1 = m.anumb_to_atom(int(f[_m]))
                    atom2 = m.anumb_to_atom(int(f[_n]))
                    atom3 = m.anumb_to_atom(int(f[_o]))
                    atom4 = m.anumb_to_atom(int(f[_p]))

                    if atom1 and atom2 and atom3 and atom4:
                        d.atom1 = atom1
                        d.atom2 = atom2
                        d.atom3 = atom3
                        d.atom4 = atom4

                        cont.append(d)
                    else:
                        raise TypeError

                elif conf['type'] == 'cmap':
                    _m = i * conf['n']
                    _n = i * conf['n'] + 1
                    _o = i * conf['n'] + 2
                    _p = i * conf['n'] + 3
                    _q = i * conf['n'] + 4
                    _r = i * conf['n'] + 5
                    _s = i * conf['n'] + 6
                    _t = i * conf['n'] + 7

                    cm = blocks.CMap()
                    atom1 = m.anumb_to_atom(int(f[_m]))
                    atom2 = m.anumb_to_atom(int(f[_n]))
                    atom3 = m.anumb_to_atom(int(f[_o]))
                    atom4 = m.anumb_to_atom(int(f[_p]))
                    atom5 = m.anumb_to_atom(int(f[_q]))
                    atom6 = m.anumb_to_atom(int(f[_r]))
                    atom7 = m.anumb_to_atom(int(f[_s]))
                    atom8 = m.anumb_to_atom(int(f[_t]))

                    if all([atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8]):
                        cm.atom1 = atom1
                        cm.atom2 = atom2
                        cm.atom3 = atom3
                        cm.atom4 = atom4
                        cm.atom5 = atom5
                        cm.atom6 = atom6
                        cm.atom7 = atom7
                        cm.atom8 = atom8

                        m.cmaps.append(cm)
                    else:
                        raise TypeError

                else:
                    raise ValueError("unknown type : '%s'" % conf['type'])

            return True

        else:
            # psffmt other than 'NAMD'
            raise NotImplementedError



if __name__ == '__main__':
    import sys, time

    ts = time.time()
    if len(sys.argv) == 2:
        p = PSFSystem(sys.argv[1], each_chain_is_molecule=True)
    elif len(sys.argv) == 3:
        p = PSFSystem(sys.argv[1], pdbfile=sys.argv[2], each_chain_is_molecule=False)
    dt = time.time() - ts

    print(p)
    print('> parsed in %4.1f seconds' % dt)






#!/usr/bin/env python

'''
    psf2top.py - A library for converting molecular topologies
    Copyright (C) 2013  Reza Salari

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


'''
This scritp automates the conversion of PSF file to GROMACS topology.

'''

import argparse, sys, os
import logging
from pytopol import charmmpar, grotop, psf
from pytopol import __version__ as version


def _setup_logging(debug_level=logging.DEBUG):
    """ Setup initial logging.

    Args:
        debug_level: the level for debugging
    Returns:
        a logging.Logger instance

    """

    lgr = logging.getLogger('mainapp')
    lgr.setLevel(debug_level)

    frmt = logging.Formatter('%(name)-30s - %(levelname)-8s - %(message)s')
    ch = logging.StreamHandler()
    ch.setLevel(debug_level)
    ch.setFormatter(frmt)

    lgr.addHandler(ch)

    return lgr


def print_note():
    note = '''
    ------------------------------------------------------------------------
    PyTopol version %s.

    Please note that this version is considered alpha and the generated
    topologies should be validated before being used in production
    simulations.

    For more info, please see: http://github.com/resal81/pytopol
    ------------------------------------------------------------------------
    '''

    print(note % version)

def main():
    print_note()

    if len(sys.argv) <= 1:
        usage = '\nUse -h for more info. Example run:\n'
        usage += '  psf2top.py -p psffile -c charmm_prm1 [charmm_prm2 ...] [-v]'
        print(usage)
        return

    # interface
    formatter = argparse.ArgumentDefaultsHelpFormatter
    p = argparse.ArgumentParser(formatter_class=formatter)
    p.add_argument('-p', type=str, help='psf file')
    p.add_argument('-c', default=[None], action='store', type=str, nargs='*',
    	help='charmm parameter files')
    p.add_argument('-v', action='store_true', help='verbose output', default=False)

    args = p.parse_args()

    # setup logging
    log_level = logging.DEBUG if args.v else logging.ERROR
    lgr = _setup_logging(log_level)
    lgr.debug('>> started main')

    # read the PSF file
    psffile = args.p
    psfsys = psf.PSFSystem(psffile)
    psfsys.split_psf()
    if len(psfsys.molecules) == 0:
        lgr.error("could not build PSF system - see above")
        return


    # read all the parameter files
    prmfiles = args.c
    par = charmmpar.CharmmPar(*prmfiles)


    # add parameters to the system
    par.add_params_to_system(psfsys, panic_on_missing_param=True)

    # convert system to gromacs format
    grotop.SystemToGroTop(psfsys)
    if not os.path.exists('top.top') or not os.path.exists('itp_mol_01.itp'):
    	lgr.error("could not build GROAMCS top file - see above")

    lgr.debug('<< exiting main \n')

    print('Done.')

main()








import logging

module_logger = logging.getLogger('mainapp.grotop')


class GroTop:
    pass


class SystemToGroTop(object):

    formats = {
        'atomtypes'    : '{:<7s} {:3d} {:>7.4f}   {:4.1f}   {:3s}     {:14.12f}     {:10.7f}  \n',
        'atoms'        : '{:6d} {:>10s} {:6d} {:6s} {:6s} {:6d} {:11.4f} {:11.4f} \n',
        'bondtypes'    : '{:5s}  {:5s}  {:1d}  {:6.4f}  {:6.1f}\n',
        'bonds'        : '{:3d}  {:3d}   {:1d}\n',
        'pairtypes'    : '{:6s} {:6s}   {:d}    {:14.12f}     {:14.12f}    \n',
        'pairs'        : '{:3d} {:3d}   {:1d}\n',
        'angletypes'   : '{:6s} {:6s} {:6s} {:1d}    {:8.4f}    {:10.5f}    {:9.5f}    {:11.5f}\n',
        'angles'       : '{:3d} {:3d} {:3d}   {:1d}\n',
        'dihedraltypes': '{:6s} {:6s} {:6s} {:6s}   {:1d}    {:6.2f}    {:10.5f}    {:d}\n',
        'dihedrals'    : '{:3d} {:3d} {:3d} {:3d}   {:1d}\n',
        'impropertypes': '{:6s} {:6s} {:6s} {:6s} {:1d} {:6.2f} {:8.4f} \n',
        'impropers'    : '{:3d} {:3d} {:3d} {:3d}   {:1d}\n',
    }


    toptemplate = ""
    toptemplate += "[ atomtypes ]      \n*ATOMTYPES*    \n"
    toptemplate += "[ nonbond_params ] \n*NONBOND_PARAM* \n"
    toptemplate += "[ pairtypes ]    \n*PAIRTYPES*    \n"
    toptemplate += "[ bondtypes ]    \n*BONDTYPES*    \n"
    toptemplate += "[ angletypes ]   \n*ANGLETYPES*   \n"
    toptemplate += "[ dihedraltypes ]\n*DIHEDRALTYPES*\n"
    toptemplate += "[ dihedraltypes ]\n*IMPROPERTYPES*\n"
    toptemplate += "[ cmaptypes ]    \n*CMAPTYPES*\n"

    itptemplate = ""
    itptemplate += "[ moleculetype ] \n*MOLECULETYPE* \n"
    itptemplate += "[ atoms ]        \n*ATOMS*        \n"
    itptemplate += "[ bonds ]        \n*BONDS*        \n"
    itptemplate += "[ pairs ]        \n*PAIRS*        \n"
    itptemplate += "[ angles ]       \n*ANGLES*       \n"
    itptemplate += "[ dihedrals ]    \n*DIHEDRALS*    \n"
    itptemplate += "[ dihedrals ]    \n*IMPROPERS*    \n"
    itptemplate += "[ cmap ]        \n*CMAPS*    \n"




    def __init__(self, psfsystem):
        self.lgr = logging.getLogger('mainapp.grotop.SystemToGroTop')
        self.lgr.debug(">> entering SystemToGroTop")

        self.system   = psfsystem
        self.assemble_topology()

        self.lgr.debug("<< leaving SystemToGroTop")


    @staticmethod
    def _redefine_atomtypes(mol):

        i = 1

        for atom in mol.atoms:
            atom.atomtype = 'at%03d' % i
            i += 1


    def assemble_topology(self, redefine_atom_types = False):

        self.lgr.debug("starting to assemble topology...")

        top = '[ defaults ] ; \n'
        top += ';nbfunc    comb-rule    gen-pairs    fudgeLJ    fudgeQQ \n'

        if self.system.forcefield == 'charmm':
            top += '1          2           yes          1.0       1.0 \n'

        self.lgr.debug("making atom/pair/bond/angle/dihedral/improper types")
        top += self.toptemplate
        top = top.replace('*ATOMTYPES*',     ''.join( self._make_atomtypes(self.system)) )
        top = top.replace('*NONBOND_PARAM*', ''.join( self._make_nonbond_param(self.system)) )
        top = top.replace('*PAIRTYPES*',     ''.join( self._make_pairtypes(self.system)) )
        top = top.replace('*BONDTYPES*',     ''.join( self._make_bondtypes(self.system)) )
        top = top.replace('*ANGLETYPES*',    ''.join( self._make_angletypes(self.system)))
        top = top.replace('*DIHEDRALTYPES*', ''.join( self._make_dihedraltypes(self.system)) )
        top = top.replace('*IMPROPERTYPES*', ''.join( self._make_impropertypes(self.system)) )
        top = top.replace('*CMAPTYPES*',     ''.join( self._make_cmaptypes(self.system)) )

        for i, m in enumerate(self.system.molecules):
            molname = 'mol_%02d' % (i+1)
            top += '#include "itp_%s.itp" \n' % molname

        top += '\n[system]  \nConvertedSystem\n\n'
        top += '[molecules] \n'

        for i, m in enumerate(self.system.molecules):
            molname = 'mol_%02d' % (i+1)
            top += '%s     1\n' % molname
        top += '\n'

        with open('top.top', 'w') as f:
            f.writelines([top])

        self.lgr.debug('writing top finished')


        self.lgr.debug("generating atom/pair/bond/angle/dihedral/improper for the itp files")


        for i,m in enumerate(self.system.molecules):
            molname = 'mol_%02d' % (i+1)
            itp = self.itptemplate
            itp = itp.replace('*MOLECULETYPE*',  ''.join( self._make_moleculetype(m, molname))  )
            itp = itp.replace('*ATOMS*',         ''.join( self._make_atoms(m))  )
            itp = itp.replace('*BONDS*',         ''.join( self._make_bonds(m))  )
            itp = itp.replace('*PAIRS*',         ''.join( self._make_pairs(m))  )
            itp = itp.replace('*ANGLES*',        ''.join( self._make_angles(m)) )
            itp = itp.replace('*DIHEDRALS*',     ''.join( self._make_dihedrals(m)) )
            itp = itp.replace('*IMPROPERS*',     ''.join( self._make_impropers(m)) )
            itp = itp.replace('*CMAPS*',         ''.join( self._make_cmaps(m)) )

            with open('itp_%s.itp' % molname, 'w') as f:
                f.writelines([itp])

        self.lgr.debug('writing %d itp files finished' % (i+1))




    def _make_atomtypes(self,m):
        def get_prot(at):
            # TODO improve this
            _protons = {'C':6, 'H':1, 'N':7, 'O':8, 'S':16, 'P':15}
            if at[0] in list(_protons.keys()):
                return _protons[at[0]]
            else:
                return 0

        result = []
        for at in m.atomtypes:
            at.convert('gromacs')
            prot = get_prot(at.atype)
            ljl  = at.gromacs['param']['ljl']
            lje  = at.gromacs['param']['lje']
            line = self.formats['atomtypes'].format(at.atype, prot, at.mass, at.charge, 'A', ljl, lje)
            result.append(line)

        return result

    def _make_nonbond_param(self, m):
        result = []
        for pr in m.interactiontypes:
            at1 = pr.atype1
            at2 = pr.atype2

            pr.convert('gromacs')
            eps = pr.gromacs['param']['lje']
            sig = pr.gromacs['param']['ljl']

            fu = 1  # TODO
            line = self.formats['pairtypes'].format(at1, at2, fu, sig, eps)
            result.append(line)

        return result

    def _make_pairtypes(self,m):

        if m.forcefield == 'charmm':
            mix_e = lambda x, y: (x*y)**0.5
            mix_l = lambda x, y: (x+y)* 0.5


        inter_at_types = {(h.atype1, h.atype2):h for h in m.interactiontypes}
        inter_keys = list(inter_at_types.keys())

        result = []
        for i in range(len(m.atomtypes)):
            m.atomtypes[i].convert('gromacs')

            for j in range(i, len(m.atomtypes)):
                m.atomtypes[j].convert('gromacs')

                at1 = m.atomtypes[i].atype
                at2 = m.atomtypes[j].atype

                if (at1, at2) in inter_keys:
                    inter_at_types[(at1,at2)].convert('gromacs')
                    e14 = inter_at_types[(at1,at2)].gromacs['param']['lje']
                    l14 = inter_at_types[(at1,at2)].gromacs['param']['ljl']
                elif (at2, at1) in inter_keys:
                    inter_at_types[(at2,at1)].convert('gromacs')
                    e14 = inter_at_types[(at2,at1)]['param']['lje']
                    l14 = inter_at_types[(at2,at1)]['param']['ljl']
                else:
                    i_lje14 = m.atomtypes[i].gromacs['param']['lje14']
                    j_lje14 = m.atomtypes[j].gromacs['param']['lje14']
                    i_ljl14 = m.atomtypes[i].gromacs['param']['ljl14']
                    j_ljl14 = m.atomtypes[j].gromacs['param']['ljl14']

                    if i_lje14 and j_lje14:
                        e14 = mix_e(i_lje14, j_lje14)
                        l14 = mix_l(i_ljl14, j_ljl14)
                    elif i_lje14:
                        j_lje = m.atomtypes[j].gromacs['param']['lje']
                        j_ljl = m.atomtypes[j].gromacs['param']['ljl']
                        e14 = mix_e(i_lje14, j_lje)
                        l14 = mix_l(i_ljl14, j_ljl)
                    elif j_lje14:
                        i_lje = m.atomtypes[i].gromacs['param']['lje']
                        i_ljl = m.atomtypes[i].gromacs['param']['ljl']
                        e14 = mix_e(i_lje, j_lje14)
                        l14 = mix_l(i_ljl, j_ljl14)
                    else:
                        continue

                fu = 1 # TODO

                line = self.formats['pairtypes'].format(at1, at2, fu, l14, e14)
                result.append(line)

        return result


    def _make_bondtypes(self,m):
        result = []
        for bond in m.bondtypes:
            at1 = bond.atype1
            at2 = bond.atype2
            bond.convert('gromacs')

            kb = bond.gromacs['param']['kb']
            b0 = bond.gromacs['param']['b0']
            fu = bond.gromacs['func']

            line = self.formats['bondtypes'].format(at1, at2, fu, b0, kb)
            result.append(line)

        return result


    def _make_angletypes(self,m):
        result = []
        for ang in m.angletypes:
            at1 = ang.atype1
            at2 = ang.atype2
            at3 = ang.atype3
            ang.convert('gromacs')

            ktetha = ang.gromacs['param']['ktetha']
            tetha0 = ang.gromacs['param']['tetha0']
            kub    = ang.gromacs['param']['kub']
            s0     = ang.gromacs['param']['s0']

            fu = ang.gromacs['func']

            line = self.formats['angletypes'].format(at1, at2, at3, fu, tetha0, ktetha, s0, kub)
            result.append(line)

        return result

    def _make_dihedraltypes(self,m):
        result = []
        for dih in m.dihedraltypes:
            at1 = dih.atype1
            at2 = dih.atype2
            at3 = dih.atype3
            at4 = dih.atype4

            dih.convert('gromacs')
            fu = dih.gromacs['func']

            for dpar in dih.gromacs['param']:
                kchi = dpar['kchi']
                n    = dpar['n']
                delta= dpar['delta']

                line = self.formats['dihedraltypes'].format(at1, at2, at3, at4, fu, delta, kchi, n)
                result.append(line)

        return result

    def _make_impropertypes(self,m):
        result = []
        for imp in m.impropertypes:
            at1 = imp.atype1
            at2 = imp.atype2
            at3 = imp.atype3
            at4 = imp.atype4
            imp.convert('gromacs')

            kpsi = imp.gromacs['param']['kpsi']
            psi0 = imp.gromacs['param']['psi0']

            fu = imp.gromacs['func']
            line = self.formats['impropertypes'].format(at1, at2, at3, at4, fu, psi0, kpsi)
            result.append(line)

        return result

    def _make_cmaptypes(self, m):
        result = []
        for cmap in m.cmaptypes:
            at1 = cmap.atype1
            at2 = cmap.atype2
            at3 = cmap.atype3
            at4 = cmap.atype4
            at5 = cmap.atype5
            at6 = cmap.atype6
            at7 = cmap.atype7
            at8 = cmap.atype8

            cmap.convert('gromacs')

            fu = cmap.gromacs['func']
            line = '%s %s %s %s %s %d 24 24' % (at1, at2, at3, at4, at8, fu)
            for i,c in enumerate(cmap.gromacs['param']):
                if i%10 == 0:
                    line += '\\\n'
                else:
                    line += ' '
                line += '%12.8f' % c

            line += '\n\n'
            result.append(line)

        return result

    def _make_moleculetype(self,m, molname):
        return ['; Name \t\t  nrexcl \n %s    3 \n' % molname]

    def _make_atoms(self,m):
        result = []
        #i = 1
        for atom in m.atoms:
            numb = cgnr = atom.number
            atype = atom.get_atomtype()
            assert atype!= False and hasattr(atom, 'charge') and hasattr(atom, 'mass')
            line = self.formats['atoms'].format(
                    numb, atype, atom.residue.number, atom.residue.name, atom.name, cgnr, atom.charge, atom.mass)
            result.append(line)

        result.insert(0,'; %5d atoms\n' % len(result))
        return result

    def _make_pairs(self,m):

        result = []
        for pr in m.pairs:
            fu = 1
            p1 = pr.atom1.number
            p4 = pr.atom2.number

            line = self.formats['pairs'].format(p1, p4, fu)
            result.append(line)

        result.insert(0,'; %5d pairs\n' % len(result))
        return result


    def _make_bonds(self,m):
        result = []
        for bond in m.bonds:
            fu = 1
            line = self.formats['bonds'].format(bond.atom1.number, bond.atom2.number, fu)
            result.append(line)

        result.insert(0,'; %5d bonds\n' % len(result))
        return result

    def _make_angles(self,m):
        result = []
        for ang in m.angles:
            fu = 5
            line = self.formats['angles'].format(ang.atom1.number, ang.atom2.number, ang.atom3.number, fu)
            result.append(line)

        result.insert(0,'; %5d angles\n' % len(result))
        return result

    def _make_dihedrals(self,m):
        result = []
        for dih in m.dihedrals:
            fu = 9
            line = self.formats['dihedrals'].format(
                    dih.atom1.number, dih.atom2.number, dih.atom3.number, dih.atom4.number, fu)
            result.append(line)

        result.insert(0,'; %5d dihedrals\n' % len(result))
        return result

    def _make_impropers(self,m):
        result = []
        for imp in m.impropers:
            fu = 2
            line = self.formats['impropers'].format(
                    imp.atom1.number, imp.atom2.number, imp.atom3.number, imp.atom4.number, fu)
            result.append(line)

        result.insert(0,'; %5d impropers\n' % len(result))
        return result

    def _make_cmaps(self, m):
        result = []

        for cmap in m.cmaps:
            fu = 1
            line = '%5d %5d %5d %5d %5d   %d\n' % (
                cmap.atom1.number, cmap.atom2.number, cmap.atom3.number, cmap.atom4.number,
                cmap.atom8.number, fu)
            result.append(line)

        result.insert(0,'; %5d cmaps\n' % len(result))
        return result













import os, logging, time
from pytopol.psf import PSFSystem
from pytopol.par import ParType
import blocks


module_logger = logging.getLogger('mainapp.charmmpar')

'''
http://www.ks.uiuc.edu/Training/Tutorials/namd/namd-tutorial-unix-html/node23.html
'''


class CharmmPar(object):
    """ A class for reading CHARMM PAR/PRM files. """

    def __init__(self, *args):
        """ Constructor.

        Args:
            One or more strings. Each string specifies path to one par file.

        Attributes:
            lgr         : logging.Logger
            fnames      : a list of paths to the par files
            bondpars    : ParType, bonds
            anglepars   : ParType, angles
            dihedralpars: ParType, dihedrals
            improperpars: ParType, impropers
            nonbonding  : ParType, nonbonding
            cmappars    : ParType, cmaps
            nbfix       : ParType, nbfix

        """

        self.lgr = logging.getLogger('mainapp.charmmpar.CharmmPar')
        self.lgr.debug(">> entering CharmmPar")


        self.fnames = args
        self.bondpars     = ParType(sym=True, mult=False, name='bond'      )
        self.anglepars    = ParType(sym=True, mult=False, name='angle'     )
        self.dihedralpars = ParType(sym=True, mult=True,  name='dihedral'  )
        self.improperpars = ParType(sym=True, mult=False, name='improper'  )
        self.nonbonding   = ParType(sym=False,mult=False, name='nonbonding')
        self.nbfix        = ParType(sym=True, mult=False, name='nbfix')
        self.cmappars     = ParType(sym=False,mult=False, name='cmap')

        for p in args:
            self._parse_charmmpar(p)
            self.lgr.debug(self.__repr__())

        self.lgr.debug("<< leaving CharmmPar")



    def __repr__(self):
        return "%3d bonding, %3d angle, %3d dih, %3d imp, %3d nonbonding and %2d cmaps" % (
                len(self.bondpars), len(self.anglepars), len(self.dihedralpars),
                len(self.improperpars), len(self.nonbonding), len(self.cmappars))


    def _parse_charmmpar(self, fname):
        """ A method for parsing CHARMM PAR/PRM files. """

        self.lgr.debug("parsing parameter file: %s" % fname)
        t1 = time.time()

        # sections of the par file (doesn't include CMAP, which will be parsed separately)
        main_parts = {
          'BOND':   {'nfields':(4,),     'nheader':2, 'cont':self.bondpars    },
          'ANGL':   {'nfields':(5,7),    'nheader':3, 'cont':self.anglepars   },
          'TETH':   {'nfields':(5,7),    'nheader':3, 'cont':self.anglepars   },
          'DIHE':   {'nfields':(7,10,13),'nheader':4, 'cont':self.dihedralpars},
          'IMPR':   {'nfields':(7,10,13),'nheader':4, 'cont':self.improperpars},
          'IMPH':   {'nfields':(7,10,13),'nheader':4, 'cont':self.improperpars},
          'NONB':   {'nfields':(4,7),    'nheader':1, 'cont':self.nonbonding  },
          'NBON':   {'nfields':(4,7),    'nheader':1, 'cont':self.nonbonding  },
          'NBFI':   {'nfields':(4,),     'nheader':2, 'cont':self.nbfix       },
         }

        # check if the file exists
        if not os.path.exists(fname):
            raise IOError("the '%s' CHARMM PAR file doesn't exist" % fname)

        # cache all of lines
        with open(fname) as f:
            _lines = f.readlines()


        # helper function to parse a line -------------------------------------
        def _parse_par_line(line, _curr_par):

            f = line.split()

            # check the number of fields
            if not len(f) in main_parts[_curr_par]['nfields']:
                return False, "section %s - number of fields didn't match: %d \n  %s" % (
                        _curr_par, len(f), line)

            if _curr_par in ('BOND',):
                if len(f) == 4:
                    at1, at2, kb, b0 = f
                    main_parts[_curr_par]['cont'].add_parameter((at1,at2), (float(kb),float(b0)))

            elif _curr_par in ('ANGL', 'TETH'):
                if len(f) == 5:
                    at1, at2, at3, ktetha, tetha0 = f
                    main_parts[_curr_par]['cont'].add_parameter(
                            (at1,at2,at3), (float(ktetha),float(tetha0), None, None) )
                elif len(f) == 7:
                    at1, at2, at3, ktetha, tetha0, kub, s0 = f
                    main_parts[_curr_par]['cont'].add_parameter(
                            (at1,at2,at3), (float(ktetha),float(tetha0), float(kub), float(s0)) )
                else:
                    raise ValueError(line)

            elif _curr_par in ('DIHE',):
                key = (f[0], f[1], f[2], f[3])
                nsets = int((len(f)-4)/3)
                for i in range(nsets):
                    kchi = float(f[4+ i*3+0])
                    n    = int  (f[4+ i*3+1])
                    delta= float(f[4+ i*3+2])
                    main_parts[_curr_par]['cont'].add_parameter(key, (kchi, n, delta) )


            elif _curr_par in ('IMPR', 'IMPH'):
                key = (f[0], f[1], f[2], f[3])
                nsets = int((len(f)-4)/3)
                for i in range(nsets):
                    kpsi = float(f[4+ i*3+0])
                    psi0 = float(f[4+ i*3+2])
                    main_parts[_curr_par]['cont'].add_parameter(key, (kpsi, psi0) )

            elif _curr_par in ('NONB', 'NBON'):
                if len(f) == 4:
                    at, tmp, epsilon, rmin2 = f
                    main_parts[_curr_par]['cont'].add_parameter(
                            at, (float(epsilon), float(rmin2), None, None) )
                elif len(f)==7:
                    at, tmp, epsilon, rmin2, tmp, epsilon14, rmin2_14 = f
                    main_parts[_curr_par]['cont'].add_parameter(
                            at, (float(epsilon), float(rmin2), float(epsilon14), float(rmin2_14)) )
                else:
                    raise ValueError(line)

            elif _curr_par in ('NBFI',):
                if len(f) == 4:
                    at1, at2, epsilon, rmin = f
                    main_parts[_curr_par]['cont'].add_parameter((at1,at2), (float(epsilon),float(rmin)))
                else:
                    raise NotImplementedError

            else:
                raise NotImplementedError

            return True, "OK"


        def _parse_cmap_lines(lines, cmappars):
            # assuming the cmap grid is 24x24

            n = 0   # should be zero for modulus
            p = []
            key = None
            for i, line in enumerate(lines):
                if n % (24*24) == 0:

                    if len(p) >0:
                        if len(p) != 24 * 24:
                            print('warning - not enough item for the cmap', key)
                        cmappars.add_parameter(key, p)

                    key = tuple(line.split()[:8])

                    p = []
                    n = 1

                else:
                    p += list(map(float, line.split()))
                    n = len(p)

            # last one
            if len(p) > 0:
                if len(p) != 24 * 24:
                    print('warning - not enough item for the cmap', key)
                cmappars.add_parameter(key, p)



        # go over the lines in the file
        _main_sections = tuple(main_parts.keys())
        _curr_par      = None

        cm_lines = [] # for caching CMAP lines

        for ln, line in enumerate(_lines):
            if '!' in line:
                line = line[0: line.index('!')]
            line = line.strip()

            if line == '':
                continue

            elif line=='END' or line=='end':
                break

            elif line.startswith('*'):
                pass

            elif line.startswith('cutnb'):
                pass

            elif line.startswith(( 'HBOND', )):
                _curr_par = None

            elif line.startswith('CMAP'):
                _curr_par = 'CMAP'

            elif line.startswith(_main_sections):
                _curr_par = line[:4]

            elif _curr_par is not None:
                # parsing normal line that has parameter data

                if _curr_par == 'CMAP':
                    cm_lines.append(line)
                else:
                    result, msg = _parse_par_line(line, _curr_par)

                    if result is False:
                        self.lgr.error(msg)
                        return False

        if len(cm_lines) > 0:
            _parse_cmap_lines(cm_lines, self.cmappars)

        t2 = time.time()
        self.lgr.debug("parsing took %4.1f seconds" % (t2-t1))






    def add_params_to_system(self, system, panic_on_missing_param=True):
        self.lgr.debug("adding parameters to the system...")
        assert isinstance(system , PSFSystem)

        added_atomtypes = []
        added_bondtypes = []
        added_angletypes = []
        added_dihedraltypes = []
        added_impropertypes = []
        added_cmaptypes = []

        system.forcefield = 'charmm'

        for mi, mol in enumerate(system.molecules):

            for atom in mol.atoms:
                at = atom.get_atomtype()
                if not at:
                    raise ValueError('atom type for atom %s was not found' % atom)

                if at in added_atomtypes:
                    continue

                added_atomtypes.append(at)
                p = self.nonbonding.get_parameter(at)

                if len(p) != 1:
                    msg = "for atom type %s, %d parameters found (expecting 1, found: %s)" % (at, len(p), p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 4
                lje, ljl, lje14, ljl14 = p[0]

                newat = blocks.AtomType('charmm')
                newat.atype = at
                newat.mass  = atom.mass
                newat.charge= atom.charge
                newat.charmm['param']['lje'] = lje
                newat.charmm['param']['ljl'] = ljl
                newat.charmm['param']['lje14'] = lje14
                newat.charmm['param']['ljl14'] = ljl14
                system.atomtypes.append(newat)



            for bond in mol.bonds:
                at1 = bond.atom1.get_atomtype()
                at2 = bond.atom2.get_atomtype()

                if (at1, at2) in added_bondtypes or (at2, at1) in added_bondtypes:
                    continue

                added_bondtypes.append((at1, at2))
                p = self.bondpars.get_parameter((at1, at2))

                if len(p) != 1:
                    msg = "for bond %s-%s, %d parameters found (expecting 1, found %s)" % (at1, at2, len(p), p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 2
                kb, b0 = p[0]

                newbt = blocks.BondType('charmm')
                newbt.atype1 = at1
                newbt.atype2 = at2
                newbt.charmm['param']['kb'] = kb
                newbt.charmm['param']['b0'] = b0
                system.bondtypes.append(newbt)



            for angle in mol.angles:
                at1 = angle.atom1.get_atomtype()
                at2 = angle.atom2.get_atomtype()
                at3 = angle.atom3.get_atomtype()

                if (at1, at2, at3) in added_angletypes or (at3, at2, at1) in added_angletypes:
                    continue

                added_angletypes.append((at1, at2, at3))
                p = self.anglepars.get_parameter((at1, at2, at3))

                if len(p) != 1:
                    msg = "for angle %s-%s-%s, %d parameters found (expecting 1, found %s)" % (
                        at1, at2, at3, len(p),p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 4
                ktetha, tetha0, kub, s0 = p[0]
                kub = kub if kub else 0
                s0  = s0  if s0  else 0

                newagt = blocks.AngleType('charmm')
                newagt.atype1 = at1
                newagt.atype2 = at2
                newagt.atype3 = at3
                newagt.charmm['param']['ktetha'] = ktetha
                newagt.charmm['param']['tetha0'] = tetha0
                newagt.charmm['param']['kub']    = kub
                newagt.charmm['param']['s0']     = s0

                system.angletypes.append(newagt)



            for dihedral in mol.dihedrals:
                at1 = dihedral.atom1.get_atomtype()
                at2 = dihedral.atom2.get_atomtype()
                at3 = dihedral.atom3.get_atomtype()
                at4 = dihedral.atom4.get_atomtype()

                if (at1, at2, at3, at4) in added_dihedraltypes or (at4, at3, at2, at1) in added_dihedraltypes:
                    continue

                added_dihedraltypes.append((at1, at2 , at3, at4))
                p = self.dihedralpars.get_charmm_dihedral_wildcard((at1, at2, at3, at4))

                if len(p) == 0:
                    msg = "for dihedral %s-%s-%s-%s no parameters was found" % (at1, at2, at3, at3)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                newdih = blocks.DihedralType('charmm')
                newdih.atype1 = at1
                newdih.atype2 = at2
                newdih.atype3 = at3
                newdih.atype4 = at4

                for dp in p:
                    kchi, n, delta = dp

                    m = {'kchi':kchi, 'n':n, 'delta': delta}
                    newdih.charmm['param'].append(m)

                system.dihedraltypes.append(newdih)



            for improper in mol.impropers:
                at1 = improper.atom1.get_atomtype()
                at2 = improper.atom2.get_atomtype()
                at3 = improper.atom3.get_atomtype()
                at4 = improper.atom4.get_atomtype()

                if (at1, at2, at3, at4) in added_impropertypes or (at4, at3, at2, at1) in added_impropertypes:
                    continue

                added_impropertypes.append((at1, at2, at3, at4))
                p = self.improperpars.get_charmm_improper_wildcard((at1, at2, at3, at4))

                if len(p) != 1:
                    msg = "for improper %s-%s-%s-%s, %d parameters found (expecting 1, found %s)" % (
                        at1, at2, at3, at4, len(p), p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 2
                kpsi, psi0 = p[0]

                newimp = blocks.ImproperType('charmm')
                newimp.atype1 = at1
                newimp.atype2 = at2
                newimp.atype3 = at3
                newimp.atype4 = at4

                newimp.charmm['param']['kpsi'] = kpsi
                newimp.charmm['param']['psi0'] = psi0

                system.impropertypes.append(newimp)



            for cmap in mol.cmaps:
                at1 = cmap.atom1.get_atomtype()
                at2 = cmap.atom2.get_atomtype()
                at3 = cmap.atom3.get_atomtype()
                at4 = cmap.atom4.get_atomtype()
                at5 = cmap.atom5.get_atomtype()
                at6 = cmap.atom6.get_atomtype()
                at7 = cmap.atom7.get_atomtype()
                at8 = cmap.atom8.get_atomtype()

                if ((at1, at2, at3, at4, at5, at6, at7, at8)) in added_cmaptypes:
                    continue

                added_cmaptypes.append((at1, at2, at3, at4, at5, at6, at7, at8))
                p = self.cmappars.get_parameter((at1, at2, at3, at4, at5, at6, at7, at8))

                if len(p) != 1:
                    msg = "for cmap: %s-%s-%s-%s-%s-%s-%s-%s, %d parameters found (expecting 1, found %s)" % (
                        at1, at2, at3, at4, at5, at6, at7, at8, len(p), p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg, len(p))
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 24*24, '%d != %d' % (len(p[0]), 24*24)
                newcmt = blocks.CMapType('charmm')
                newcmt.atype1 = at1
                newcmt.atype2 = at2
                newcmt.atype3 = at3
                newcmt.atype4 = at4
                newcmt.atype5 = at5
                newcmt.atype6 = at6
                newcmt.atype7 = at7
                newcmt.atype8 = at8
                newcmt.charmm['param'] = p[0]

                system.cmaptypes.append(newcmt)


            for nbkey in list(self.nbfix._data.keys()):
                if mi != 0:
                    continue   # just for the first molecule

                at1, at2 = nbkey
                p = self.nbfix.get_parameter((at1, at2))
                if len(p) == 0:
                    continue  #

                print('adding pair param to the mol for %s-%s' % (at1,at2))

                if len(p) != 1:
                    msg = "for pair %s-%s, %d parameters found (expecting 1, found %s)" % (at1, at2, len(p), p)
                    if not panic_on_missing_param:
                        self.lgr.error(msg)
                        continue
                    else:
                        raise ValueError(msg)

                assert len(p[0]) == 2
                eps, rmin = p[0]
                newnb = blocks.InteractionType('charmm')
                newnb.atype1 = at1
                newnb.atype2 = at2
                newnb.charmm['param']['lje'] = eps
                newnb.charmm['param']['ljl'] = rmin

                system.interactiontypes.append(newnb)

if __name__ == '__main__':
    import sys
    c = CharmmPar(sys.argv[1])
    print(c)








class System(object):
    def __init__(self):
        self.molecules = tuple([])

        self.atomtypes = []
        self.bondtypes = []
        self.angletypes= []
        self.dihedraltypes    = []
        self.impropertypes    = []
        self.cmaptypes        = []
        self.interactiontypes = []
        self.forcefield=  None



class Molecule(object):

    def __init__(self):
        self.chains    = []
        self.atoms     = []
        self.residues  = []

        self.bonds     = []
        self.angles    = []
        self.dihedrals = []
        self.impropers = []
        self.cmaps     = []
        self.pairs     = []

        self._anumb_to_atom = {}


    def anumb_to_atom(self, anumb):
        '''Returns the atom object corresponding to an atom number'''

        assert isinstance(anumb, int), "anumb must be integer"

        if len(self._anumb_to_atom) == 0:   # empty dictionary

            if len(self.atoms) != 0:
                for atom in self.atoms:
                    self._anumb_to_atom[atom.number] = atom
                return self._anumb_to_atom[anumb]
            else:
                print("no atoms in the molecule")
                return False

        else:
            if anumb in self._anumb_to_atom:
                return self._anumb_to_atom[anumb]
            else:
                print("no such atom number (%d) in the molecule" % (anumb))
                return False


    def renumber_atoms(self):

        if len(self.atoms) != 0:

            # reset the mapping
            self._anumb_to_atom = {}

            for i,atom in enumerate(self.atoms):
                atom.number = i+1   # starting from 1

        else:
            print("the number of atoms is zero - no renumbering")


class Chain(object):
    """
        name    = str,
        residues= list,
        molecule= Molecule
    """

    def __init__(self):
        self.residues = []


class Residue(object):
    """
        name    = str,
        number  = int,
        chain   = Chain,
        chain_name = str,
        atoms   = list,
    """

    def __init__(self):
        self.atoms  = []


class Atom(object):
    """
        name    = str,
        number  = int,
        flag    = str,        # HETATM
        coords  = list,
        residue = Residue,
        occup   = float,
        bfactor = float,
        altlocs = list,
        atomtype= str,
        radius  = float,
        charge  = radius,
        mass    = float,
        chain   = str,
        resname = str,
        resnumb = int,
        altloc  = str,         # per atoms

    """

    def __init__(self):

        self.coords = []        # a list of coordinates (x,y,z) of models
        self.altlocs= []        # a list of (altloc_name, (x,y,z), occup, bfactor)




    def get_atomtype(self):
        if hasattr(self, 'atomtype'):
            return self.atomtype
        else:
            print("atom %s doesn't have atomtype" % self)
            return False



class Bond:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None

class Angle:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None
        self.atom3 = None

class Dihedral:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None
        self.atom3 = None
        self.atom4 = None

class Improper:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None
        self.atom3 = None
        self.atom4 = None

class CMap:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None
        self.atom3 = None
        self.atom4 = None
        self.atom5 = None
        self.atom6 = None
        self.atom7 = None
        self.atom8 = None

class Pair:
    def __init__(self):
        self.atom1 = None
        self.atom2 = None



class Param:
    def convert(self, reqformat):
        assert reqformat in ('charmm', 'gromacs')

        if reqformat == self.format:
            if reqformat == 'charmm':
                return self.charmm
            elif reqformat == 'gromacs':
                return self.gromacs
            else:
                raise NotImplementedError



        if isinstance(self, AtomType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']['lje'] = abs(self.charmm['param']['lje']) * 4.184
                self.gromacs['param']['ljl'] = self.charmm['param']['ljl'] * 2 * 0.1 / (2**(1.0/6.0))

                if self.charmm['param']['lje14'] is not None:
                    self.gromacs['param']['lje14'] = abs(self.charmm['param']['lje14']) * 4.184
                    self.gromacs['param']['ljl14'] = self.charmm['param']['ljl14'] * 2 * 0.1 / (2**(1.0/6.0))
                else:
                    self.gromacs['param']['lje14'] = None
                    self.gromacs['param']['ljl14'] = None
            else:
                raise NotImplementedError



        elif isinstance(self, BondType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']['kb'] = self.charmm['param']['kb'] * 2 * 4.184 * (1.0 / 0.01)   # nm^2
                self.gromacs['param']['b0'] = self.charmm['param']['b0'] * 0.1
                self.gromacs['func'] = 1
            else:
                raise NotImplementedError



        elif isinstance(self, AngleType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']['ktetha'] = self.charmm['param']['ktetha'] * 2 * 4.184
                self.gromacs['param']['tetha0'] = self.charmm['param']['tetha0']
                self.gromacs['param']['kub'] = self.charmm['param']['kub'] * 2 * 4.184 * 10 * 10
                self.gromacs['param']['s0'] = self.charmm['param']['s0'] * 0.1
                self.gromacs['func'] = 5
            else:
                raise NotImplementedError



        elif isinstance(self, DihedralType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                for dih in self.charmm['param']:
                    convdih = {}
                    convdih['kchi']  = dih['kchi'] * 4.184
                    convdih['n']     = dih['n']
                    convdih['delta'] = dih['delta']
                    self.gromacs['param'].append(convdih)
                self.gromacs['func'] = 9
            else:
                raise NotImplementedError



        elif isinstance(self, ImproperType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']['kpsi'] = self.charmm['param']['kpsi'] * 2 * 4.184
                self.gromacs['param']['psi0'] = self.charmm['param']['psi0']
                self.gromacs['func'] = 2
            else:
                raise NotImplementedError



        elif isinstance(self, CMapType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']= [n*4.184 for n in self.charmm['param']]
                self.gromacs['func'] = 1
            else:
                raise NotImplementedError



        elif isinstance(self, InteractionType):
            if reqformat == 'gromacs' and self.format == 'charmm':
                self.gromacs['param']['lje'] = abs(self.charmm['param']['lje']) * 4.184
                self.gromacs['param']['ljl'] = self.charmm['param']['ljl'] * 0.1 /  (2**(1.0/6.0))   # no *2

                if self.charmm['param']['lje14'] is not None:
                    self.gromacs['param']['lje14'] = abs(self.charmm['param']['lje14']) * 4.184
                    self.gromacs['param']['ljl14'] = self.charmm['param']['ljl14'] * 0.1 / (2**(1.0/6.0))
                else:
                    self.gromacs['param']['lje14'] = None
                    self.gromacs['param']['ljl14'] = None
            else:
                raise NotImplementedError

        else:
            raise NotImplementedError


class AtomType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format
        self.atype  = None
        self.mass   = None
        self.charge = None

        self.charmm = {'param': {'lje':None, 'ljl':None, 'lje14':None, 'ljl14':None} }
        self.gromacs= {'param': {'lje':None, 'ljl':None, 'lje14':None, 'ljl14':None} }


class BondType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format
        self.atype1 = None
        self.atype2 = None

        self.charmm = {'param': {'kb':None, 'b0':None} }
        self.gromacs= {'param': {}, 'func':None}


class AngleType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format

        self.atype1 = None
        self.atype2 = None
        self.atype3 = None

        self.charmm = {'param':{'ktetha':None, 'tetha0':None, 'kub':None, 's0':None} }
        self.gromacs= {'param':{}}


class DihedralType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format

        self.atype1 = None
        self.atype2 = None
        self.atype3 = None
        self.atype4 = None

        self.charmm = {'param':[]}  # {kchi, n, delta}
        self.gromacs= {'param':[]}


class ImproperType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format

        self.atype1 = None
        self.atype2 = None
        self.atype3 = None
        self.atype4 = None

        self.charmm = {'param': {'kpsi': None, 'psi0':None} }
        self.gromacs= {'param':{}, 'func':None}


class CMapType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format

        self.atype1 = None
        self.atype2 = None
        self.atype3 = None
        self.atype4 = None
        self.atype5 = None
        self.atype6 = None
        self.atype7 = None
        self.atype8 = None

        self.charmm = {'param': []}
        self.gromacs= {'param': []}


class InteractionType(Param):
    def __init__(self, format):
        assert format in ('charmm', 'gromacs')
        self.format = format

        self.atype1 = None
        self.atype2 = None

        self.charmm = {'param': {'lje':None, 'ljl':None, 'lje14':None, 'ljl14':None} }
        self.gromacs= {'param': {'lje':None, 'ljl':None, 'lje14':None, 'ljl14':None} }






from collections import defaultdict
import logging

module_logger = logging.getLogger('mainapp.par')

class ParType(object):
    '''
    For CHARMM, the self.data is a dictionary with the following format:

        - bonds
                !V(bond) = Kb(b - b0)**2
                !
                !Kb: kcal/mole/A**2
                !b0: A
                !
                !atom type Kb          b0

            self.data = { (atype1, atype2) : [ (Kb, b0 ) ] }

        - angles
                !V(angle) = Ktheta(Theta - Theta0)**2
                !
                !V(Urey-Bradley) = Kub(S - S0)**2
                !
                !Ktheta: kcal/mole/rad**2
                !Theta0: degrees
                !Kub: kcal/mole/A**2 (Urey-Bradley)
                !S0: A
                !
                !atom types     Ktheta    Theta0   Kub     S0

            self.data = { (atype1, atype2, atype3) : [ (Ktheta, Theta0, Kub or None, S0 or None) ]


        - dihedrals
                !V(dihedral) = Kchi(1 + cos(n(chi) - delta))
                !
                !Kchi: kcal/mole
                !n: multiplicity
                !delta: degrees
                !
                !atom types             Kchi    n   delta

            self.data = { (atype1, atype2, atype3, atype4) : [ (Kchi, n, delta) ]


        - impropers
                !V(improper) = Kpsi(psi - psi0)**2
                !
                !Kpsi: kcal/mole/rad**2
                !psi0: degrees
                !note that the second column of numbers (0) is ignored
                !
                !atom types           Kpsi    ignored        psi0

            self.data = { (atype1, atype2, atype3, atype4) : [ (Kpsi, psi0) ]


        - nonbonding
                !V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]
                !
                !epsilon: kcal/mole, Eps,i,j = sqrt(eps,i * eps,j)
                !Rmin/2: A, Rmin,i,j = Rmin/2,i + Rmin/2,j
                !
                !atom  ignored    epsilon      Rmin/2   ignored   eps,1-4       Rmin/2,1-4

            self.data = { atype : [ (epsilon, Rmin/2, eps14 or None, Rmin14 or None) ] }

        - nbfix
                !               Emin        Rmin
                !            (kcal/mol)     (A)

            self.data = {(atype1, atype2) : [ (epsilon, Rmin)]}

    '''



    def __init__(self, sym=False, mult=False, name=''):
        self.symmetric_keys         = sym
        self.multiple_value_per_key = mult

        assert name in ['bond', 'angle', 'dihedral', 'improper', 'nonbonding', 'cmap', 'nbfix']
        self.name = name

        self._data = defaultdict(list)      # (atype1, atype2) : [(coeffs)]
        self._key_set = set()

        self.lgr = logging.getLogger('mainapp.par.Par')


    def __len__(self):
        return len(self._data)


    def add_parameter(self, key, value):
        if self.multiple_value_per_key:
            # no check - append new values
            self._data[key].append(value)
        else:
            # no mutliple values - maximum one value per key - either write or overwrite
            key1 = key
            key2 = key[::-1] if (isinstance(key, tuple) and self.symmetric_keys) else key

            old_keys = list(self._data.keys())

            if not key1 in old_keys and not key2 in old_keys:
                self._data[key1].append(value)
            else:
                key = key1 if key1 in old_keys else key2
                if self._data[key][0] != value:
                    self.lgr.warning('overwritten: %s -> key: %s' % ( self.name, key))
                    self.lgr.warning('  from: %s, to: %s' % (self._data[key], value))
                    self._data[key][0] = value


    def get_parameter(self, key):
        if len(self._key_set) == 0 and len(list(self._data.keys())) > 0:
            self._key_set = set(self._data.keys())


        if key in self._key_set:
            return self._data[key]
        else:
            if isinstance(key, tuple) and self.symmetric_keys:
                if key[::-1] in self._key_set:
                    return self._data[key[::-1]]

        return []



    def get_charmm_dihedral_wildcard(self, key):
        # X-atomtype2-atomtype3-X
        result = self.get_parameter(key)
        if result != []:
            return result
        else:
            return self.get_parameter( ('X', key[1], key[2], 'X') )


    def get_charmm_improper_wildcard(self,key):
        # atomtype1-X-X-atomtype4
        result = self.get_parameter(key)
        if result != []:
            return result
        else:
            return self.get_parameter( (key[0], 'X', 'X', key[3]) )





from pytopol import psf
from config import psf_files as ref


# set up the systems
psf_systems = {}
for name in list(ref.keys()):
    path = ref[name]['path']
    psf_systems[name] = psf.PSFSystem(path, each_chain_is_molecule=True)


def test_parsing_psf_files():
    assert psf_systems != {}
    for name in list(psf_systems.keys()):
        assert isinstance(psf_systems[name].molecules, tuple)
        assert len(psf_systems[name].molecules) == ref[name]['nmolecules']

def test_nmolecules():
    for name in list(psf_systems.keys()):
        assert len(psf_systems[name].molecules) == ref[name]['nmolecules']

def test_natoms():
    for name in list(psf_systems.keys()):
        natoms = sum([len(m.atoms) for m in psf_systems[name].molecules])
        assert natoms == ref[name]['natoms']

def test_nresidues():
    for name in list(psf_systems.keys()):
        nresidues = sum([len(m.residues) for m in psf_systems[name].molecules])
        assert nresidues == ref[name]['nresidues']

def test_nchains():
    for name in list(psf_systems.keys()):
        nchains = sum([len(m.chains) for m in psf_systems[name].molecules])
        assert nchains == ref[name]['nchains']

def test_nbonds():
    for name in list(psf_systems.keys()):
        nbonds = sum([len(m.bonds) for m in psf_systems[name].molecules])
        assert nbonds == ref[name]['nbonds']

def test_nangles():
    for name in list(psf_systems.keys()):
        nangles = sum([len(m.angles) for m in psf_systems[name].molecules])
        assert nangles == ref[name]['nangles']

def test_ndihedrals():
    for name in list(psf_systems.keys()):
        ndihedrals = sum([len(m.dihedrals) for m in psf_systems[name].molecules])
        assert ndihedrals == ref[name]['ndihedrals']

def test_nimpropers():
    for name in list(psf_systems.keys()):
        nimpropers = sum([len(m.impropers) for m in psf_systems[name].molecules])
        assert nimpropers == ref[name]['nimpropers']






#!/usr/bin/env python

import os
import sys
from setuptools.command.test import test as TestCommand
import pytopol

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the
        #eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


#readme = open('README.md').read()
doclink = """
Documentation
=============

Plese see http://github.com/resal81/pytopol"""
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='pytopol',
    version=pytopol.__version__,
    description='"Library for converting molecular topologies"',
    long_description= doclink + '\n\n' + history,
    author='Reza Salari',
    author_email='rezasalari@rutgers.edu',
    url='https://github.com/resal81/pytopol',
    packages=[
        'pytopol',
    ],
    package_dir={'pytopol': 'pytopol'},
    include_package_data=True,
    install_requires=[
    ],
    license='GPLv3',
    zip_safe=False,
    keywords='pytopol',
    scripts=['scripts/psf2top.py'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    tests_require=['pytest>=2.3.5'],
    cmdclass={'test': PyTest},
)




from pytopol import blocks
from pytopol.utils import build_res_chain
import os, logging, time



module_logger = logging.getLogger('mainapp.pdb')



class PDBSystem(object):

    def __init__(self, pdbfile, guess_mols=False):
        """ PDB parser.

        Attributes:
            pdbfile : str, path to the pdb file
            molecules: tuple Molecules
            atoms : tuple of Atoms
            lgt : logging.Logger

        """


        self.lgr = logging.getLogger('mainapp.pdb.PDBSystem')
        self.lgr.debug(">> entering PDBSystem")

        self.pdbfile = pdbfile
        self.atoms   = tuple([])
        self.molecules = tuple([])

        self._parse(self.pdbfile, guess_mols)

        self.lgr.debug("<< leaving PDBsystem")

    def _parse(self, pdbfile, guess_mols):
        self.lgr.debug("parsing pdb file: %s" % pdbfile)

        t1 = time.time()

        if not os.path.exists(pdbfile):
            self.lgr.error("the pdbfile doesn't exist")
            return

        _first_model_finished = False
        atoms     = []
        molecules = []
        _i = 0   # a counter for atom index
        _alt_loc_warning = False

        # read the file and create atoms list
        M = blocks.Molecule()
        molecules.append(M)
        with open(pdbfile) as f:
            for line in f:
                line = line.strip()
                if line.startswith('ENDMDL'):
                    _first_model_finished = True   # set by first ENDMDL
                    _i = 0   # reset _i

                if _first_model_finished:
                    # just read the coordinates
                    if line.startswith(('ATOM', 'HETATM')):
                        c = list(map(float, (line[30:38], line[38:46], line[46:54])))
                        atoms[_i].coords.append(c)
                        _i += 1
                else:
                    if line.startswith(('ATOM', 'HETATM')):
                        a = blocks.Atom()
                        a.flag   = line[0:6].strip()
                        a.number = self.conv_atom_number(line[6:11])
                        a.name   = line[12:16].strip()
                        a.altloc = line[16].strip()
                        if a.altloc != '' and _alt_loc_warning is False:
                            _alt_loc_warning = True
                        a.resname= line[17:21].strip()
                        a.chain  = line[21].strip()
                        a.resnumb= int(line[22:26])
                        c = list(map(float, (line[30:38], line[38:46], line[46:54])))
                        a.coords = [tuple(c)]  # a list of (x,y,z) tuples, each tuple for one model

                        #TODO occup, bfactor, ...

                        if guess_mols:
                            if len(atoms) > 0:
                                if line.startswith('HETATM'):
                                    if a.resname != atoms[-1].resname or a.resnumb != atoms[-1].resnumb:
                                        M = blocks.Molecule()
                                        molecules.append(M)
                                else: # ATOM
                                    if atoms[-1].flag == 'HETATM':
                                        M = blocks.Molecule()
                                        molecules.append(M)

                        # record a in the local atoms and M.atoms
                        atoms.append(a)
                        M.atoms.append(a)

        if len(atoms) == 0:
            self.lgr.warning("no atoms were found in the pdb file")
            return

        self.atoms = tuple(atoms)

        if _alt_loc_warning:
            self.lgr.warning("there are atom records with altloc flags - fix this")

        # make sure all the atomic coordinates are the same length
        for a in atoms:
            assert len(a.coords) == len(atoms[0].coords)

        # build residue and chains
        for m in molecules:
            build_res_chain(m)

        self.molecules = tuple(molecules)


        t2 = time.time()
        self.lgr.debug("parsing took %4.1f seconds" % (t2-t1))

    @staticmethod
    def conv_atom_number(s):
        try:
            numb = int(s)
        except ValueError:
            numb = int(s, 16)
        return numb

    def __repr__(self):
        natoms = sum([len(m.atoms) for m in self.molecules])
        return "PDBSystem with %6d atoms and %2d molecules" % (natoms, len(self.molecules))







if __name__ == '__main__':
    import sys, time

    t1 = time.time()
    p = PDBSystem(sys.argv[1])
    dt = time.time() - t1
    print(p)
    print('> parsed in %4.1f seconds' % dt)








from pytopol import charmmpar
from config import par_files as ref

# set up the charmmpars
pars = {}
for name in list(ref.keys()):
    path = ref[name]['path']
    pars[name] = charmmpar.CharmmPar(path)


def test_nbonds():
	for name in list(ref.keys()):
		assert len(pars[name].bondpars) == ref[name]['nbonds']


def test_nangles():
	for name in list(ref.keys()):
		assert len(pars[name].anglepars) == ref[name]['nangles']

def test_dihedrals():
	for name in list(ref.keys()):
		ndih = sum([len(v) for v in list(pars[name].dihedralpars._data.values())])
		assert ndih == ref[name]['ndihedrals']

def test_nimpropers():
	for name in list(ref.keys()):
		assert len(pars[name].improperpars) == ref[name]['nimpropers']

def test_nnonbonding():
	for name in list(ref.keys()):
		assert len(pars[name].nonbonding) == ref[name]['nnonbonding']

def test_ncmaps():
	for name in list(ref.keys()):
		assert len(pars[name].cmappars) == ref[name]['ncmaps']




import logging
from pytopol import blocks

lgr = logging.getLogger('mainapp.utils')


def build_res_chain(m):
    # using a molecule object with atoms, builds residues and chains
    R = None
    residues = []
    for i, a in enumerate(m.atoms):
        if R is None or (a.resname  != R.name or
                          a.resnumb != R.number or
                          a.chain   != R.chain_name):
            R = blocks.Residue()
            R.name = a.resname
            R.number = a.resnumb
            R.chain_name  = a.chain
            residues.append(R)

        R.atoms.append(a)
        a.residue = R

    m.residues = residues

    # chains
    C = None   # current chain object
    chains = []
    for i, r in enumerate(m.residues):
        if C is None or (r.chain_name != C.name):
            C = blocks.Chain()
            C.name = r.chain_name
            chains.append(C)

        C.residues.append(r)
        r.chain = C

    m.chains = chains

def build_pairs(m):
    # using a molecule with bonds, angles and dihedrals, build pairs
    print('building pairs with %d bonds, %d angles and %d dihedrals' % (
        len(m.bonds), len(m.angles), len(m.dihedrals)))

    _bonds = []
    for bond in m.bonds:
        _bonds.append((bond.atom1.number, bond.atom2.number))

    _angles = []
    for ang in m.angles:
        _angles.append((ang.atom1.number, ang.atom3.number))

    _bonds = set(_bonds)
    _angles = set(_angles)

    _pairs = set([])
    for dih in m.dihedrals:
        p1 = dih.atom1.number
        p4 = dih.atom4.number

        if (p1,p4) in _bonds or (p1,p4) in _angles or \
           (p4,p1) in _bonds or (p4,p1) in _angles:
            continue

        if (p1,p4) in _pairs or (p4,p1) in _pairs:
            continue

        _pairs.add((p1,p4))

        thispair = blocks.Pair()
        thispair.atom1 = dih.atom1
        thispair.atom2 = dih.atom4

        m.pairs.append(thispair)







